"""Stub generator for C modules.

The public interface is via the mypy.stubgen module.
"""

import importlib
import os.path
import re
from typing import List, Dict, Tuple, Optional
from types import ModuleType

from mypy.stubutil import (
    parse_all_signatures, find_unique_signatures, is_c_module, write_header,
    infer_sig_from_docstring
)


def generate_stub_for_c_module(module_name: str,
                               target: str,
                               add_header: bool = True,
                               sigs: Dict[str, str] = {},
                               class_sigs: Dict[str, str] = {},
                               ) -> None:
    module = importlib.import_module(module_name)
    assert is_c_module(module), '%s is not a C module' % module_name
    subdir = os.path.dirname(target)
    if subdir and not os.path.isdir(subdir):
        os.makedirs(subdir)
    functions = []  # type: List[str]
    done = set()
    items = sorted(module.__dict__.items(), key=lambda x: x[0])
    for name, obj in items:
        if is_c_function(obj):
            generate_c_function_stub(module, name, obj, functions, sigs=sigs)
            done.add(name)
    types = []  # type: List[str]
    for name, obj in items:
        if name.startswith('__') and name.endswith('__'):
            continue
        if is_c_type(obj):
            generate_c_type_stub(module, name, obj, types, sigs=sigs, class_sigs=class_sigs)
            done.add(name)
    variables = []
    for name, obj in items:
        if name.startswith('__') and name.endswith('__'):
            continue
        if name not in done:
            type_str = type(obj).__name__
            if type_str not in ('int', 'str', 'bytes', 'float', 'bool'):
                type_str = 'Any'
            variables.append('%s = ... # type: %s' % (name, type_str))
    output = []
    for line in variables:
        output.append(line)
    if output and functions:
        output.append('')
    for line in functions:
        output.append(line)
    for line in types:
        if line.startswith('class') and output and output[-1]:
            output.append('')
        output.append(line)
    output = add_typing_import(output)
    with open(target, 'w') as file:
        if add_header:
            write_header(file, module_name)
        for line in output:
            file.write('%s\n' % line)


def add_typing_import(output: List[str]) -> List[str]:
    names = []
    for name in ['Any']:
        if any(re.search(r'\b%s\b' % name, line) for line in output):
            names.append(name)
    if names:
        return ['from typing import %s' % ', '.join(names), ''] + output
    else:
        return output[:]


def is_c_function(obj: object) -> bool:
    return type(obj) is type(ord)


def is_c_method(obj: object) -> bool:
    return type(obj) in (type(str.index),
                         type(str.__add__),
                         type(str.__new__))


def is_c_classmethod(obj: object) -> bool:
    type_str = type(obj).__name__
    return type_str == 'classmethod_descriptor'


def is_c_type(obj: object) -> bool:
    return type(obj) is type(int)


def generate_c_function_stub(module: ModuleType,
                             name: str,
                             obj: object,
                             output: List[str],
                             self_var: Optional[str] = None,
                             sigs: Dict[str, str] = {},
                             class_name: Optional[str] = None,
                             class_sigs: Dict[str, str] = {},
                             ) -> None:
    if self_var:
        self_arg = '%s, ' % self_var
    else:
        self_arg = ''
    if (name in ('__new__', '__init__') and name not in sigs and class_name and
            class_name in class_sigs):
        sig = class_sigs[class_name]
    else:
        docstr = getattr(obj, '__doc__', None)
        inferred = infer_sig_from_docstring(docstr, name)
        if inferred:
            sig = inferred
        else:
            if class_name and name not in sigs:
                sig = infer_method_sig(name)
            else:
                sig = sigs.get(name, '(*args, **kwargs)')
    sig = sig[1:-1]
    if sig:
        if sig.split(',', 1)[0] == self_var:
            self_arg = ''
    else:
        self_arg = self_arg.replace(', ', '')
    output.append('def %s(%s%s): ...' % (name, self_arg, sig))


def generate_c_type_stub(module: ModuleType,
                         class_name: str,
                         obj: type,
                         output: List[str],
                         sigs: Dict[str, str] = {},
                         class_sigs: Dict[str, str] = {},
                         ) -> None:
    items = sorted(obj.__dict__.items(), key=lambda x: method_name_sort_key(x[0]))
    methods = []
    done = set()
    for attr, value in items:
        if is_c_method(value) or is_c_classmethod(value):
            done.add(attr)
            if not is_skipped_attribute(attr):
                if is_c_classmethod(value):
                    methods.append('@classmethod')
                    self_var = 'cls'
                else:
                    self_var = 'self'
                if attr == '__new__':
                    # TODO: We should support __new__.
                    if '__init__' in obj.__dict__:
                        # Avoid duplicate functions if both are present.
                        # But is there any case where .__new__() has a
                        # better signature than __init__() ?
                        continue
                    attr = '__init__'
                generate_c_function_stub(module, attr, value, methods, self_var, sigs=sigs,
                                         class_name=class_name, class_sigs=class_sigs)
    variables = []
    for attr, value in items:
        if is_skipped_attribute(attr):
            continue
        if attr not in done:
            variables.append('%s = ... # type: Any' % attr)
    all_bases = obj.mro()
    if all_bases[-1] is object:
        # TODO: Is this always object?
        del all_bases[-1]
    # remove the class itself
    all_bases = all_bases[1:]
    # Remove base classes of other bases as redundant.
    bases = []  # type: List[type]
    for base in all_bases:
        if not any(issubclass(b, base) for b in bases):
            bases.append(base)
    if bases:
        bases_str = '(%s)' % ', '.join(base.__name__ for base in bases)
    else:
        bases_str = ''
    if not methods and not variables:
        output.append('class %s%s: ...' % (class_name, bases_str))
    else:
        output.append('class %s%s:' % (class_name, bases_str))
        for variable in variables:
            output.append('    %s' % variable)
        for method in methods:
            output.append('    %s' % method)


def method_name_sort_key(name: str) -> Tuple[int, str]:
    if name in ('__new__', '__init__'):
        return (0, name)
    if name.startswith('__') and name.endswith('__'):
        return (2, name)
    return (1, name)


def is_skipped_attribute(attr: str) -> bool:
    return attr in ('__getattribute__',
                    '__str__',
                    '__repr__',
                    '__doc__',
                    '__dict__',
                    '__module__',
                    '__weakref__')  # For pickling


def infer_method_sig(name: str) -> str:
    if name.startswith('__') and name.endswith('__'):
        name = name[2:-2]
        if name in ('hash', 'iter', 'next', 'sizeof', 'copy', 'deepcopy', 'reduce', 'getinitargs',
                    'int', 'float', 'trunc', 'complex', 'bool'):
            return '()'
        if name == 'getitem':
            return '(index)'
        if name == 'setitem':
            return '(index, object)'
        if name in ('delattr', 'getattr'):
            return '(name)'
        if name == 'setattr':
            return '(name, value)'
        if name == 'getstate':
            return '()'
        if name == 'setstate':
            return '(state)'
        if name in ('eq', 'ne', 'lt', 'le', 'gt', 'ge',
                    'add', 'radd', 'sub', 'rsub', 'mul', 'rmul',
                    'mod', 'rmod', 'floordiv', 'rfloordiv', 'truediv', 'rtruediv',
                    'divmod', 'rdivmod', 'pow', 'rpow'):
            return '(other)'
        if name in ('neg', 'pos'):
            return '()'
    return '(*args, **kwargs)'



from typing import Any

class KazooException(Exception): ...
class ZookeeperError(KazooException): ...
class CancelledError(KazooException): ...
class ConfigurationError(KazooException): ...
class ZookeeperStoppedError(KazooException): ...
class ConnectionDropped(KazooException): ...
class LockTimeout(KazooException): ...
class WriterNotClosedException(KazooException): ...

EXCEPTIONS = ...  # type: Any

class RolledBackError(ZookeeperError): ...
class SystemZookeeperError(ZookeeperError): ...
class RuntimeInconsistency(ZookeeperError): ...
class DataInconsistency(ZookeeperError): ...
class ConnectionLoss(ZookeeperError): ...
class MarshallingError(ZookeeperError): ...
class UnimplementedError(ZookeeperError): ...
class OperationTimeoutError(ZookeeperError): ...
class BadArgumentsError(ZookeeperError): ...
class NewConfigNoQuorumError(ZookeeperError): ...
class ReconfigInProcessError(ZookeeperError): ...
class APIError(ZookeeperError): ...
class NoNodeError(ZookeeperError): ...
class NoAuthError(ZookeeperError): ...
class BadVersionError(ZookeeperError): ...
class NoChildrenForEphemeralsError(ZookeeperError): ...
class NodeExistsError(ZookeeperError): ...
class NotEmptyError(ZookeeperError): ...
class SessionExpiredError(ZookeeperError): ...
class InvalidCallbackError(ZookeeperError): ...
class InvalidACLError(ZookeeperError): ...
class AuthFailedError(ZookeeperError): ...
class SessionMovedError(ZookeeperError): ...
class NotReadOnlyCallError(ZookeeperError): ...
class ConnectionClosedError(SessionExpiredError): ...

ConnectionLossException = ...  # type: Any
MarshallingErrorException = ...  # type: Any
SystemErrorException = ...  # type: Any
RuntimeInconsistencyException = ...  # type: Any
DataInconsistencyException = ...  # type: Any
UnimplementedException = ...  # type: Any
OperationTimeoutException = ...  # type: Any
BadArgumentsException = ...  # type: Any
ApiErrorException = ...  # type: Any
NoNodeException = ...  # type: Any
NoAuthException = ...  # type: Any
BadVersionException = ...  # type: Any
NoChildrenForEphemeralsException = ...  # type: Any
NodeExistsException = ...  # type: Any
InvalidACLException = ...  # type: Any
AuthFailedException = ...  # type: Any
NotEmptyException = ...  # type: Any
SessionExpiredException = ...  # type: Any
InvalidCallbackException = ...  # type: Any



from typing import Any

class WSGIWarning(Warning): ...

def validator(application): ...

class InputWrapper:
    input = ...  # type: Any
    def __init__(self, wsgi_input): ...
    def read(self, *args): ...
    def readline(self, *args): ...
    def readlines(self, *args): ...
    def __iter__(self): ...
    def close(self): ...

class ErrorWrapper:
    errors = ...  # type: Any
    def __init__(self, wsgi_errors): ...
    def write(self, s): ...
    def flush(self): ...
    def writelines(self, seq): ...
    def close(self): ...

class WriteWrapper:
    writer = ...  # type: Any
    def __init__(self, wsgi_writer): ...
    def __call__(self, s): ...

class PartialIteratorWrapper:
    iterator = ...  # type: Any
    def __init__(self, wsgi_iterator): ...
    def __iter__(self): ...

class IteratorWrapper:
    original_iterator = ...  # type: Any
    iterator = ...  # type: Any
    closed = ...  # type: Any
    check_start_response = ...  # type: Any
    def __init__(self, wsgi_iterator, check_start_response): ...
    def __iter__(self): ...
    def __next__(self): ...
    def close(self): ...
    def __del__(self): ...



# Stubs for email.iterators (Python 3.4)

from typing import Iterator, Optional
from email.message import Message

def body_line_iterator(msg: Message, decode: bool = ...) -> Iterator[str]: ...
def typed_subpart_iterator(msg: Message, maintype: str = ...,
                           subtype: Optional[str] = ...) -> Iterator[str]: ...



# Stubs for math
# See: http://docs.python.org/2/library/math.html

from typing import Tuple, Iterable, Optional

import sys

e = ...  # type: float
pi = ...  # type: float
if sys.version_info >= (3, 5):
    inf = ...  # type: float
    nan = ...  # type: float
if sys.version_info >= (3, 6):
    tau = ...  # type: float

def acos(x: float) -> float: ...
def acosh(x: float) -> float: ...
def asin(x: float) -> float: ...
def asinh(x: float) -> float: ...
def atan(x: float) -> float: ...
def atan2(y: float, x: float) -> float: ...
def atanh(x: float) -> float: ...
def ceil(x: float) -> int: ...
def copysign(x: float, y: float) -> float: ...
def cos(x: float) -> float: ...
def cosh(x: float) -> float: ...
def degrees(x: float) -> float: ...
def erf(x: float) -> float: ...
def erfc(x: float) -> float: ...
def exp(x: float) -> float: ...
def expm1(x: float) -> float: ...
def fabs(x: float) -> float: ...
def factorial(x: int) -> int: ...
if sys.version_info >= (3,):
    def floor(x: float) -> int: ...
else:
    def floor(x: float) -> float: ...
def fmod(x: float, y: float) -> float: ...
def frexp(x: float) -> Tuple[float, int]: ...
def fsum(iterable: Iterable) -> float: ...
def gamma(x: float) -> float: ...
if sys.version_info >= (3, 5):
    def gcd(a: int, b: int) -> int: ...
def hypot(x: float, y: float) -> float: ...
if sys.version_info >= (3, 5):
    def isclose(a: float, b: float, rel_tol: float = ..., abs_tol: float = ...) -> bool: ...
def isinf(x: float) -> bool: ...
if sys.version_info >= (3,):
    def isfinite(x: float) -> bool: ...
def isnan(x: float) -> bool: ...
def ldexp(x: float, i: int) -> float: ...
def lgamma(x: float) -> float: ...
def log(x: float, base: float = ...) -> float: ...
def log10(x: float) -> float: ...
def log1p(x: float) -> float: ...
if sys.version_info >= (3, 3):
    def log2(x: float) -> float: ...
def modf(x: float) -> Tuple[float, float]: ...
def pow(x: float, y: float) -> float: ...
def radians(x: float) -> float: ...
def sin(x: float) -> float: ...
def sinh(x: float) -> float: ...
def sqrt(x: float) -> float: ...
def tan(x: float) -> float: ...
def tanh(x: float) -> float: ...
def trunc(x: float) -> int: ...



from typing import Any

UTF8_COOKIE = ...  # type: Any
system_exceptions = ...  # type: Any
HEADER = ...  # type: Any
FOOTER = ...  # type: Any
PAGE_HTML = ...  # type: Any
CONSOLE_HTML = ...  # type: Any
SUMMARY_HTML = ...  # type: Any
FRAME_HTML = ...  # type: Any
SOURCE_LINE_HTML = ...  # type: Any

def render_console_html(secret, evalex_trusted=True): ...
def get_current_traceback(ignore_system_exceptions=False, show_hidden_frames=False, skip=0): ...

class Line:
    lineno = ...  # type: Any
    code = ...  # type: Any
    in_frame = ...  # type: Any
    current = ...  # type: Any
    def __init__(self, lineno, code): ...
    def classes(self): ...
    def render(self): ...

class Traceback:
    exc_type = ...  # type: Any
    exc_value = ...  # type: Any
    exception_type = ...  # type: Any
    frames = ...  # type: Any
    def __init__(self, exc_type, exc_value, tb): ...
    def filter_hidden_frames(self): ...
    def is_syntax_error(self): ...
    def exception(self): ...
    def log(self, logfile=None): ...
    def paste(self): ...
    def render_summary(self, include_title=True): ...
    def render_full(self, evalex=False, secret=None, evalex_trusted=True): ...
    def generate_plaintext_traceback(self): ...
    def plaintext(self): ...
    id = ...  # type: Any

class Frame:
    lineno = ...  # type: Any
    function_name = ...  # type: Any
    locals = ...  # type: Any
    globals = ...  # type: Any
    filename = ...  # type: Any
    module = ...  # type: Any
    loader = ...  # type: Any
    code = ...  # type: Any
    hide = ...  # type: Any
    info = ...  # type: Any
    def __init__(self, exc_type, exc_value, tb): ...
    def render(self): ...
    def render_line_context(self): ...
    def get_annotated_lines(self): ...
    def eval(self, code, mode=''): ...
    def sourcelines(self): ...
    def get_context_lines(self, context=5): ...
    @property
    def current_line(self): ...
    def console(self): ...
    id = ...  # type: Any



from typing import Any

greenlet = ...  # type: Any

class IterIO:
    def __new__(cls, obj, sentinel=''): ...
    def __iter__(self): ...
    def tell(self): ...
    def isatty(self): ...
    def seek(self, pos, mode=0): ...
    def truncate(self, size=None): ...
    def write(self, s): ...
    def writelines(self, list): ...
    def read(self, n=-1): ...
    def readlines(self, sizehint=0): ...
    def readline(self, length=None): ...
    def flush(self): ...
    def __next__(self): ...

class IterI(IterIO):
    def __new__(cls, func, sentinel=''): ...
    closed = ...  # type: Any
    def close(self): ...
    def write(self, s): ...
    def writelines(self, list): ...
    def flush(self): ...

class IterO(IterIO):
    sentinel = ...  # type: Any
    closed = ...  # type: Any
    pos = ...  # type: Any
    def __new__(cls, gen, sentinel=''): ...
    def __iter__(self): ...
    def close(self): ...
    def seek(self, pos, mode=0): ...
    def read(self, n=-1): ...
    def readline(self, length=None): ...
    def readlines(self, sizehint=0): ...



from typing import Any

NO = ...  # type: Any
YES = ...  # type: Any
TRUE = ...  # type: Any
FALSE = ...  # type: Any
ON = ...  # type: Any
OFF = ...  # type: Any
N = ...  # type: Any
S = ...  # type: Any
W = ...  # type: Any
E = ...  # type: Any
NW = ...  # type: Any
SW = ...  # type: Any
NE = ...  # type: Any
SE = ...  # type: Any
NS = ...  # type: Any
EW = ...  # type: Any
NSEW = ...  # type: Any
CENTER = ...  # type: Any
NONE = ...  # type: Any
X = ...  # type: Any
Y = ...  # type: Any
BOTH = ...  # type: Any
LEFT = ...  # type: Any
TOP = ...  # type: Any
RIGHT = ...  # type: Any
BOTTOM = ...  # type: Any
RAISED = ...  # type: Any
SUNKEN = ...  # type: Any
FLAT = ...  # type: Any
RIDGE = ...  # type: Any
GROOVE = ...  # type: Any
SOLID = ...  # type: Any
HORIZONTAL = ...  # type: Any
VERTICAL = ...  # type: Any
NUMERIC = ...  # type: Any
CHAR = ...  # type: Any
WORD = ...  # type: Any
BASELINE = ...  # type: Any
INSIDE = ...  # type: Any
OUTSIDE = ...  # type: Any
SEL = ...  # type: Any
SEL_FIRST = ...  # type: Any
SEL_LAST = ...  # type: Any
END = ...  # type: Any
INSERT = ...  # type: Any
CURRENT = ...  # type: Any
ANCHOR = ...  # type: Any
ALL = ...  # type: Any
NORMAL = ...  # type: Any
DISABLED = ...  # type: Any
ACTIVE = ...  # type: Any
HIDDEN = ...  # type: Any
CASCADE = ...  # type: Any
CHECKBUTTON = ...  # type: Any
COMMAND = ...  # type: Any
RADIOBUTTON = ...  # type: Any
SEPARATOR = ...  # type: Any
SINGLE = ...  # type: Any
BROWSE = ...  # type: Any
MULTIPLE = ...  # type: Any
EXTENDED = ...  # type: Any
DOTBOX = ...  # type: Any
UNDERLINE = ...  # type: Any
PIESLICE = ...  # type: Any
CHORD = ...  # type: Any
ARC = ...  # type: Any
FIRST = ...  # type: Any
LAST = ...  # type: Any
BUTT = ...  # type: Any
PROJECTING = ...  # type: Any
ROUND = ...  # type: Any
BEVEL = ...  # type: Any
MITER = ...  # type: Any
MOVETO = ...  # type: Any
SCROLL = ...  # type: Any
UNITS = ...  # type: Any
PAGES = ...  # type: Any



from typing import TypeVar, Generic, Any, Iterable, Iterator, Callable, Tuple, Optional, Set, NamedTuple

FIRST_COMPLETED = ...  # type: str
FIRST_EXCEPTION = ...  # type: str
ALL_COMPLETED = ...  # type: str
PENDING = ...  # type: Any
RUNNING = ...  # type: Any
CANCELLED = ...  # type: Any
CANCELLED_AND_NOTIFIED = ...  # type: Any
FINISHED = ...  # type: Any
LOGGER = ...  # type: Any

class Error(Exception): ...
class CancelledError(Error): ...
class TimeoutError(Error): ...

DoneAndNotDoneFutures = NamedTuple('DoneAndNotDoneFutures', [('done', Future[Any]), ('not_done', Future[Any])])

_T = TypeVar('_T')

class Future(Generic[_T]):
    def __init__(self) -> None: ...
    def cancel(self) -> bool: ...
    def cancelled(self) -> bool: ...
    def running(self) -> bool: ...
    def done(self) -> bool: ...
    def add_done_callback(self, fn: Callable[[Future[_T]], Any]) -> None: ...
    def result(self, timeout: Optional[float] = ...) -> _T: ...
    def exception(self, timeout: Optional[float] = ...) -> Optional[BaseException]: ...
    def set_running_or_notify_cancel(self) -> bool: ...
    def set_result(self, result: _T) -> None: ...
    def set_exception(self, exception: Optional[BaseException]) -> None: ...

class Executor:
    def submit(self, fn: Callable[..., _T], *args: Any, **kwargs: Any) -> Future[_T]: ...
    def map(self, func: Callable[..., _T], *iterables: Iterable[Any], timeout: Optional[float] = ..., chunksize: int = ...) -> Iterator[_T]: ...
    def shutdown(self, wait: bool = ...) -> None: ...
    def __enter__(self: _T) -> _T: ...
    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> bool: ...

def as_completed(fs: Iterable[Future[_T]], timeout: Optional[float] = ...) -> Iterator[Future[_T]]: ...

def wait(fs: Iterable[Future[_T]], timeout: Optional[float] = ..., return_when: str = ...) -> Tuple[Set[Future[_T]], Set[Future[_T]]]: ...



# Stubs for types
# Note, all classes "defined" here require special handling.

# TODO parts of this should be conditional on version

import sys
from typing import (
    Any, Awaitable, Callable, Dict, Generic, Iterator, Mapping, Optional, Tuple, TypeVar,
    Union, overload, Type
)

# ModuleType is exported from this module, but for circular import
# reasons exists in its own stub file (with ModuleSpec and Loader).
from _importlib_modulespec import ModuleType as ModuleType  # Exported

_T = TypeVar('_T')
_T_co = TypeVar('_T_co', covariant=True)
_T_contra = TypeVar('_T_contra', contravariant=True)
_KT = TypeVar('_KT')
_VT = TypeVar('_VT')

class _Cell:
    cell_contents = ...  # type: Any

class FunctionType:
    __closure__ = ...  # type: Optional[Tuple[_Cell, ...]]
    __code__ = ...  # type: CodeType
    __defaults__ = ...  # type: Optional[Tuple[Any, ...]]
    __dict__ = ...  # type: Dict[str, Any]
    __globals__ = ...  # type: Dict[str, Any]
    __name__ = ...  # type: str
    __qualname__ = ...  # type: str
    __annotations__ = ...  # type: Dict[str, Any]
    __kwdefaults__ = ...  # type: Dict[str, Any]
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
    def __get__(self, obj: Optional[object], type: Optional[type]) -> 'MethodType': ...
LambdaType = FunctionType

class CodeType:
    """Create a code object.  Not for the faint of heart."""
    co_argcount = ...  # type: int
    co_kwonlyargcount = ...  # type: int
    co_nlocals = ...  # type: int
    co_stacksize = ...  # type: int
    co_flags = ...  # type: int
    co_code = ...  # type: bytes
    co_consts = ...  # type: Tuple[Any, ...]
    co_names = ...  # type: Tuple[str, ...]
    co_varnames = ...  # type: Tuple[str, ...]
    co_filename = ...  # type: Optional[str]
    co_name = ...  # type: str
    co_firstlineno = ...  # type: int
    co_lnotab = ...  # type: bytes
    co_freevars = ...  # type: Tuple[str, ...]
    co_cellvars = ...  # type: Tuple[str, ...]
    def __init__(
        self,
        argcount: int,
        kwonlyargcount: int,
        nlocals: int,
        stacksize: int,
        flags: int,
        codestring: bytes,
        constants: Tuple[Any, ...],
        names: Tuple[str, ...],
        varnames: Tuple[str, ...],
        filename: str,
        name: str,
        firstlineno: int,
        lnotab: bytes,
        freevars: Tuple[str, ...] = ...,
        cellvars: Tuple[str, ...] = ...,
    ) -> None: ...

class MappingProxyType(Mapping[_KT, _VT], Generic[_KT, _VT]):
    def __init__(self, mapping: Mapping[_KT, _VT]) -> None: ...
    def __getitem__(self, k: _KT) -> _VT: ...
    def __iter__(self) -> Iterator[_KT]: ...
    def __len__(self) -> int: ...

# TODO: use __getattr__ and __setattr__ instead of inheriting from Any, pending mypy#521.
class SimpleNamespace(Any): ...  # type: ignore

class GeneratorType:
    gi_code = ...  # type: CodeType
    gi_frame = ...  # type: FrameType
    gi_running = ...  # type: bool
    gi_yieldfrom = ...  # type: Optional[GeneratorType]
    def __iter__(self) -> 'GeneratorType': ...
    def __next__(self) -> Any: ...
    def close(self) -> None: ...
    def send(self, arg: Any) -> Any: ...
    @overload
    def throw(self, val: BaseException) -> Any: ...
    @overload
    def throw(self, typ: type, val: BaseException = ..., tb: 'TracebackType' = ...) -> Any: ...

if sys.version_info >= (3, 6):
    class AsyncGeneratorType(Generic[_T_co, _T_contra]):
        ag_await: Optional[Awaitable[Any]]
        ag_frame: FrameType
        ag_running: bool
        ag_code: CodeType
        def __aiter__(self) -> Awaitable[AsyncGeneratorType[_T_co, _T_contra]]: ...
        def __anext__(self) -> Awaitable[_T_co]: ...
        def asend(self, val: _T_contra) -> Awaitable[_T_co]: ...
        @overload
        def athrow(self, val: BaseException) -> Awaitable[_T_co]: ...
        @overload
        def athrow(self, typ: Type[BaseException], val: BaseException, tb: TracebackType = ...) -> Awaitable[_T_co]: ...
        def aclose(self) -> Awaitable[_T_co]: ...

class CoroutineType:
    cr_await = ...  # type: Optional[Any]
    cr_code = ...  # type: CodeType
    cr_frame = ...  # type: FrameType
    cr_running = ...  # type: bool
    def close(self) -> None: ...
    def send(self, arg: Any) -> Any: ...
    @overload
    def throw(self, val: BaseException) -> Any: ...
    @overload
    def throw(self, typ: type, val: BaseException = ..., tb: 'TracebackType' = ...) -> Any: ...

class _StaticFunctionType:
    """Fictional type to correct the type of MethodType.__func__.

    FunctionType is a descriptor, so mypy follows the descriptor protocol and
    converts MethodType.__func__ back to MethodType (the return type of
    FunctionType.__get__). But this is actually a special case; MethodType is
    implemented in C and its attribute access doesn't go through
    __getattribute__.

    By wrapping FunctionType in _StaticFunctionType, we get the right result;
    similar to wrapping a function in staticmethod() at runtime to prevent it
    being bound as a method.
    """
    def __get__(self, obj: Optional[object], type: Optional[type]) -> 'FunctionType': ...

class MethodType:
    __func__ = ...  # type: _StaticFunctionType
    __self__ = ...  # type: object
    __name__ = ...  # type: str
    __qualname__ = ...  # type: str
    def __init__(self, func: Callable, obj: object) -> None: ...
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
class BuiltinFunctionType:
    __self__ = ...  # type: Union[object, ModuleType]
    __name__ = ...  # type: str
    __qualname__ = ...  # type: str
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
BuiltinMethodType = BuiltinFunctionType

class TracebackType:
    tb_frame = ...  # type: FrameType
    tb_lasti = ...  # type: int
    tb_lineno = ...  # type: int
    tb_next = ...  # type: TracebackType

class FrameType:
    f_back = ...  # type: FrameType
    f_builtins = ...  # type: Dict[str, Any]
    f_code = ...  # type: CodeType
    f_globals = ...  # type: Dict[str, Any]
    f_lasti = ...  # type: int
    f_lineno = ...  # type: int
    f_locals = ...  # type: Dict[str, Any]
    f_trace = ...  # type: Callable[[], None]

    def clear(self) -> None: ...

class GetSetDescriptorType:
    __name__ = ...  # type: str
    __objclass__ = ...  # type: type
    def __get__(self, obj: Any, type: type = ...) -> Any: ...
    def __set__(self, obj: Any) -> None: ...
    def __delete__(self, obj: Any) -> None: ...
class MemberDescriptorType:
    __name__ = ...  # type: str
    __objclass__ = ...  # type: type
    def __get__(self, obj: Any, type: type = ...) -> Any: ...
    def __set__(self, obj: Any) -> None: ...
    def __delete__(self, obj: Any) -> None: ...

def new_class(name: str, bases: Tuple[type, ...] = ..., kwds: Dict[str, Any] = ..., exec_body: Callable[[Dict[str, Any]], None] = ...) -> type: ...
def prepare_class(name: str, bases: Tuple[type, ...] = ..., kwds: Dict[str, Any] = ...) -> Tuple[type, Dict[str, Any], Dict[str, Any]]: ...

# Actually a different type, but `property` is special and we want that too.
DynamicClassAttribute = property

def coroutine(f: Callable[..., Any]) -> CoroutineType: ...



"""Stubs for the 'sys' module."""

from typing import (
    IO, Union, List, Sequence, Any, Dict, Tuple, BinaryIO, Optional, Callable,
    overload, Text, Type,
)
from types import FrameType, ModuleType, TracebackType, ClassType
from mypy_extensions import NoReturn

class _flags:
    bytes_warning = ...  # type: int
    debug = ...  # type: int
    division_new = ...  # type: int
    division_warning = ...  # type: int
    dont_write_bytecode = ...  # type: int
    hash_randomization = ...  # type: int
    ignore_environment = ...  # type: int
    inspect = ...  # type: int
    interactive = ...  # type: int
    no_site = ...  # type: int
    no_user_site = ...  # type: int
    optimize = ...  # type: int
    py3k_warning = ...  # type: int
    tabcheck = ...  # type: int
    unicode = ...  # type: int
    verbose = ...  # type: int

class _float_info:
    max = ...  # type: float
    max_exp = ...  # type: int
    max_10_exp = ...  # type: int
    min = ...  # type: float
    min_exp = ...  # type: int
    min_10_exp = ...  # type: int
    dig = ...  # type: int
    mant_dig = ...  # type: int
    epsilon = ...  # type: float
    radix = ...  # type: int
    rounds = ...  # type: int

class _version_info(Tuple[int, int, int, str, int]):
    major = 0
    minor = 0
    micro = 0
    releaselevel = ...  # type: str
    serial = 0

_mercurial = ...  # type: Tuple[str, str, str]
api_version = ...  # type: int
argv = ...  # type: List[str]
builtin_module_names = ...  # type: Tuple[str, ...]
byteorder = ...  # type: str
copyright = ...  # type: str
dont_write_bytecode = ...  # type: bool
exec_prefix = ...  # type: str
executable = ...  # type: str
flags = ...  # type: _flags
float_repr_style = ...  # type: str
hexversion = ...  # type: int
long_info = ...  # type: object
maxint = ...  # type: int
maxsize = ...  # type: int
maxunicode = ...  # type: int
modules = ...  # type: Dict[str, Any]
path = ...  # type: List[str]
platform = ...  # type: str
prefix = ...  # type: str
py3kwarning = ...  # type: bool
__stderr__ = ...  # type: IO[str]
__stdin__ = ...  # type: IO[str]
__stdout__ = ...  # type: IO[str]
stderr = ...  # type: IO[str]
stdin = ...  # type: IO[str]
stdout = ...  # type: IO[str]
subversion = ...  # type: Tuple[str, str, str]
version = ...  # type: str
warnoptions = ...  # type: object
float_info = ...  # type: _float_info
version_info = ...  # type: _version_info
ps1 = ...  # type: str
ps2 = ...  # type: str
last_type = ...  # type: type
last_value = ...  # type: BaseException
last_traceback = ...  # type: TracebackType
# TODO precise types
meta_path = ...  # type: List[Any]
path_hooks = ...  # type: List[Any]
path_importer_cache = ...  # type: Dict[str, Any]
displayhook = ...  # type: Optional[Callable[[int], None]]
excepthook = ...  # type: Optional[Callable[[type, BaseException, TracebackType], None]]
exc_type = ...  # type: Optional[type]
exc_value = ...  # type: Union[BaseException, ClassType]
exc_traceback = ...  # type: TracebackType

class _WindowsVersionType:
    major = ...  # type: Any
    minor = ...  # type: Any
    build = ...  # type: Any
    platform = ...  # type: Any
    service_pack = ...  # type: Any
    service_pack_major = ...  # type: Any
    service_pack_minor = ...  # type: Any
    suite_mask = ...  # type: Any
    product_type = ...  # type: Any

def getwindowsversion() -> _WindowsVersionType: ...

def _clear_type_cache() -> None: ...
def _current_frames() -> Dict[int, FrameType]: ...
def _getframe(depth: int = ...) -> FrameType: ...
def call_tracing(fn: Any, args: Any) -> Any: ...
def __displayhook__(value: int) -> None: ...
def __excepthook__(type_: type, value: BaseException, traceback: TracebackType) -> None: ...
def exc_clear() -> None:
    raise DeprecationWarning()
# TODO should be a union of tuple, see mypy#1178
def exc_info() -> Tuple[Optional[Type[BaseException]],
                        Optional[BaseException],
                        Optional[TracebackType]]: ...

# sys.exit() accepts an optional argument of anything printable
def exit(arg: Any = ...) -> NoReturn:
    raise SystemExit()
def getcheckinterval() -> int: ...  # deprecated
def getdefaultencoding() -> str: ...
def getdlopenflags() -> int: ...
def getfilesystemencoding() -> str: ...  # In practice, never returns None
def getrefcount(arg: Any) -> int: ...
def getrecursionlimit() -> int: ...
def getsizeof(obj: object, default: int = ...) -> int: ...
def getprofile() -> None: ...
def gettrace() -> None: ...
def setcheckinterval(interval: int) -> None: ...  # deprecated
def setdlopenflags(n: int) -> None: ...
def setdefaultencoding(encoding: Text) -> None: ...  # only exists after reload(sys)
def setprofile(profilefunc: Any) -> None: ...  # TODO type
def setrecursionlimit(limit: int) -> None: ...
def settrace(tracefunc: Any) -> None: ...  # TODO type



from typing import Any

def dumps(*args): ...
def render_template(name_parts, rules, converters): ...
def generate_map(map, name=''): ...
def generate_adapter(adapter, name='', map_name=''): ...
def js_to_url_function(converter): ...
def NumberConverter_js_to_url(conv): ...

js_to_url_functions = ...  # type: Any



from typing import Any, Iterator, MutableMapping, Tuple, TypeVar, Generic

_VT = TypeVar('_VT')

class CaseInsensitiveDict(MutableMapping[str, _VT], Generic[_VT]):
    def lower_items(self) -> Iterator[Tuple[str, _VT]]: ...
    def __setitem__(self, key: str, value: _VT) -> None: ...
    def __getitem__(self, key: str) -> _VT: ...
    def __delitem__(self, key: str) -> None: ...
    def __iter__(self) -> Iterator[str]: ...
    def __len__(self) -> int: ...

class LookupDict(dict):
    name = ...  # type: Any
    def __init__(self, name=...) -> None: ...
    def __getitem__(self, key): ...
    def get(self, key, default=...): ...



# Stubs for tempfile
# Ron Murawski <ron@horizonchess.com>

# based on http://docs.python.org/3.3/library/tempfile.html

import sys
from types import TracebackType
from typing import Any, AnyStr, Generic, IO, Optional, Tuple, Type

# global variables
TMP_MAX: int
tempdir = ...  # type: Optional[str]
template = ...  # type: str


if sys.version_info >= (3, 5):
    def TemporaryFile(
        mode: str = ..., buffering: int = ..., encoding: str = ...,
        newline: str = ..., suffix: Optional[AnyStr]= ..., prefix: Optional[AnyStr] = ...,
        dir: Optional[AnyStr] = ...
    ) -> IO[Any]:
        ...
    def NamedTemporaryFile(
        mode: str = ..., buffering: int = ..., encoding: str = ...,
        newline: str = ..., suffix: Optional[AnyStr] = ..., prefix: Optional[AnyStr] = ...,
        dir: Optional[AnyStr] = ..., delete: bool =...
    ) -> IO[Any]:
        ...
    def SpooledTemporaryFile(
        max_size: int = ..., mode: str = ..., buffering: int = ...,
        encoding: str = ..., newline: str = ..., suffix: Optional[AnyStr] = ...,
        prefix: Optional[AnyStr] = ..., dir: Optional[AnyStr] = ...
    ) -> IO[Any]:
        ...

    class TemporaryDirectory(Generic[AnyStr]):
        name = ...  # type: str
        def __init__(self, suffix: Optional[AnyStr] = ..., prefix: Optional[AnyStr] = ...,
                     dir: Optional[AnyStr] = ...) -> None: ...
        def cleanup(self) -> None: ...
        def __enter__(self) -> AnyStr: ...
        def __exit__(self, exc_type: Optional[Type[BaseException]],
                     exc_val: Optional[BaseException],
                     exc_tb: Optional[TracebackType]) -> bool: ...

    def mkstemp(suffix: Optional[AnyStr] = ..., prefix: Optional[AnyStr] = ..., dir: Optional[AnyStr] = ...,
                text: bool = ...) -> Tuple[int, AnyStr]: ...
    def mkdtemp(suffix: Optional[AnyStr] = ..., prefix: Optional[AnyStr] = ...,
                dir: Optional[str] = ...) -> AnyStr: ...
    def mktemp(suffix: Optional[AnyStr] = ..., prefix: Optional[AnyStr] = ..., dir: Optional[AnyStr] = ...) -> AnyStr: ...

    def gettempdirb() -> bytes: ...
    def gettempprefixb() -> bytes: ...
else:
    def TemporaryFile(
        mode: str = ..., buffering: int = ..., encoding: str = ...,
        newline: str = ..., suffix: str = ..., prefix: str = ...,
        dir: Optional[str] = ...
    ) -> IO[Any]:
        ...
    def NamedTemporaryFile(
        mode: str = ..., buffering: int = ..., encoding: str = ...,
        newline: str = ..., suffix: str = ..., prefix: str = ...,
        dir: Optional[str] = ..., delete: bool =...
    ) -> IO[Any]:
        ...
    def SpooledTemporaryFile(
        max_size: int = ..., mode: str = ..., buffering: int = ...,
        encoding: str = ..., newline: str = ..., suffix: str = ...,
        prefix: str = ..., dir: Optional[str] = ...
    ) -> IO[Any]:
        ...

    class TemporaryDirectory:
        name = ...  # type: str
        def __init__(self, suffix: str = ..., prefix: str = ...,
                     dir: Optional[str] = ...) -> None: ...
        def cleanup(self) -> None: ...
        def __enter__(self) -> str: ...
        def __exit__(self, exc_type: Optional[Type[BaseException]],
                     exc_val: Optional[BaseException],
                     exc_tb: Optional[TracebackType]) -> bool: ...

    def mkstemp(suffix: str = ..., prefix: str = ..., dir: Optional[str] = ...,
                text: bool = ...) -> Tuple[int, str]: ...
    def mkdtemp(suffix: str = ..., prefix: str = ...,
                dir: Optional[str] = ...) -> str: ...
    def mktemp(suffix: str = ..., prefix: str = ..., dir: Optional[str] = ...) -> str: ...

def gettempdir() -> str: ...
def gettempprefix() -> str: ...



# Stubs for zipfile

from typing import Callable, IO, List, Optional, Tuple, Type, Union
from types import TracebackType
import sys


_SZI = Union[str, ZipInfo]
_DT = Tuple[int, int, int, int, int, int]


if sys.version_info >= (3,):
    class BadZipFile(Exception): ...
    BadZipfile = BadZipFile
else:
    class BadZipfile(Exception): ...
error = BadZipfile

class LargeZipFile(Exception): ...

class ZipFile:
    debug = ...  # type: int
    comment = ...  # type: bytes
    filelist = ...  # type: List[ZipInfo]
    def __init__(self, file: Union[str, IO[bytes]], mode: str = ..., compression: int = ...,
                 allowZip64: bool = ...) -> None: ...
    def __enter__(self) -> ZipFile: ...
    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[Exception],
                 exc_tb: Optional[TracebackType]) -> bool: ...
    def close(self) -> None: ...
    def getinfo(self, name: str) -> ZipInfo: ...
    def infolist(self) -> List[ZipInfo]: ...
    def namelist(self) -> List[str]: ...
    def open(self, name: _SZI, mode: str = ...,
             pwd: Optional[bytes] = ...) -> IO[bytes]: ...
    def extract(self, member: _SZI, path: Optional[_SZI] = ...,
                pwd: bytes = ...) -> str: ...
    def extractall(self, path: Optional[str] = ...,
                   members: Optional[List[str]] = ...,
                   pwd: Optional[bytes] = ...) -> None: ...
    def printdir(self) -> None: ...
    def setpassword(self, pwd: bytes) -> None: ...
    def read(self, name: _SZI, pwd: Optional[bytes] = ...) -> bytes: ...
    def testzip(self) -> Optional[str]: ...
    def write(self, filename: str, arcname: Optional[str] = ...,
              compress_type: Optional[int] = ...) -> None: ...
    if sys.version_info >= (3,):
        def writestr(self, zinfo_or_arcname: _SZI, data: Union[bytes, str],
                     compress_type: Optional[int] = ...) -> None: ...
    else:
        def writestr(self,
                     zinfo_or_arcname: _SZI, bytes: bytes,
                     compress_type: Optional[int] = ...) -> None: ...

class PyZipFile(ZipFile):
    if sys.version_info >= (3,):
        def __init__(self, file: Union[str, IO[bytes]], mode: str = ...,
                     compression: int = ..., allowZip64: bool = ...,
                     opimize: int = ...) -> None: ...
        def writepy(self, pathname: str, basename: str = ...,
                    filterfunc: Optional[Callable[[str], bool]] = ...) -> None: ...
    else:
        def writepy(self,
                    pathname: str, basename: str = ...) -> None: ...

class ZipInfo:
    filename = ...  # type: str
    date_time = ...  # type: _DT
    compress_type = ...  # type: int
    comment = ...  # type: bytes
    extra = ...  # type: bytes
    create_system = ...  # type: int
    create_version = ...  # type: int
    extract_version = ...  # type: int
    reserved = ...  # type: int
    flag_bits = ...  # type: int
    volume = ...  # type: int
    internal_attr = ...  # type: int
    external_attr = ...  # type: int
    header_offset = ...  # type: int
    CRC = ...  # type: int
    compress_size = ...  # type: int
    file_size = ...  # type: int
    if sys.version_info < (3,):
        def __init__(self, filename: Optional[str] = ...,
                     date_time: Optional[_DT] = ...) -> None: ...


def is_zipfile(filename: Union[str, IO[bytes]]) -> bool: ...

ZIP_STORED = ...  # type: int
ZIP_DEFLATED = ...  # type: int
if sys.version_info >= (3, 3):
    ZIP_BZIP2 = ...  # type: int
    ZIP_LZMA = ...  # type: int
